/* Copyright 2022, Michele Santullo
 * This file is part of duckticker.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with duckticker.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "bitcoinity_ticker_price.hpp"
#include <wrenpp/vm_fun.hpp>
#include <wrenpp/def_configuration.hpp>
#include <wrenpp/callback_manager.hpp>
#include <utility>
#include <algorithm>

namespace {
constexpr std::string_view g_dummy_currency {"EUR"};
constexpr std::string_view g_dummy_exchange {"kraken"};

constexpr char g_script[] =
R"script(
import "bitcoinity" for Bitcoinity

class App {
	construct new(exchange, currency) {
		_currency = currency
		_exchange = exchange
	}

	start() {
		var price_quick = Bitcoinity.ticker_quick()
		System.print("Ticker (quick): %(price_quick.price) %(price_quick.currency) on %(price_quick.exchange)")

		var price_currency = Bitcoinity.ticker_currency(_currency)
		System.print("Ticker (%(_currency)): %(price_currency.price) %(price_currency.currency) on %(price_currency.exchange)")

		var price = Bitcoinity.ticker(_exchange, _currency)
		System.print("Ticker (%(_exchange), %(_currency)): %(price.price) %(price.currency) on %(price.exchange)")
	}
}

var the_app = App.new("kraken", "USD")
)script";

void quick_ticker_price_bitcoinity_wren (wren::VM& vm) {
	using duck::TickerPrice;
	using duck::bitcoinity_ticker_price;
	using duck::BitcoinityMatchType;

	TickerPrice* const tp = wren::make_foreign_object<TickerPrice>(vm, "ticker_price");
	*tp = bitcoinity_ticker_price(g_dummy_exchange, g_dummy_currency, BitcoinityMatchType::Any);
}

void ticker_price_bitcoinity_wren (wren::VM& vm) {
	using duck::TickerPrice;
	using duck::bitcoinity_ticker_price;
	using duck::BitcoinityMatchType;

	TickerPrice* const tp = wren::make_foreign_object<TickerPrice>(vm, "ticker_price");
	*tp = bitcoinity_ticker_price(wren::get<const char*>(vm, 1), wren::get<const char*>(vm, 2), BitcoinityMatchType::Exact);
}

void currency_ticker_price_bitcoinity_wren (wren::VM& vm) {
	using duck::TickerPrice;
	using duck::bitcoinity_ticker_price;
	using duck::BitcoinityMatchType;

	TickerPrice* const tp = wren::make_foreign_object<TickerPrice>(vm, "ticker_price");
	*tp = bitcoinity_ticker_price(g_dummy_exchange, wren::get<const char*>(vm, 1), BitcoinityMatchType::CurrencyOnly);
}

void exchange_ticker_price_bitcoinity_wren (wren::VM& vm) {
	using duck::TickerPrice;
	using duck::bitcoinity_ticker_price;
	using duck::BitcoinityMatchType;

	TickerPrice* const tp = wren::make_foreign_object<TickerPrice>(vm, "ticker_price");
	*tp = bitcoinity_ticker_price(wren::get<const char*>(vm, 1), g_dummy_currency, BitcoinityMatchType::ExchangeOnly);
}

class MyWrenConfiguration : public wren::DefConfiguration {
public:
	wren::foreign_class_t foreign_class_fn(
		wren::VM* vm,
		std::string_view module,
		std::string_view class_name
	) {
		if (module == "ticker_price" and class_name == "TickerPrice")
			return wren::make_foreign_class<duck::TickerPrice>();
		else
			return {nullptr, nullptr};
	}

	char* load_module_fn(wren::VM* vm, std::string_view module_name) {
		std::string_view wren_code;

		if (module_name == "bitcoinity")
			wren_code = duck::g_bitcoinity_wren_module;
		else if (module_name == "ticker_price")
			wren_code = duck::g_ticker_price_wren_module;
		else
			return nullptr;

		const std::size_t buff_sz = wren_code.size();
		char* const buff = static_cast<char*>(MyWrenConfiguration::reallocate_fn(nullptr, buff_sz));
		std::copy(wren_code.cbegin(), wren_code.cend(), buff);
		return buff;
	}
};
} //unnamed namespace

int main() {
	using duck::TickerPrice;

	MyWrenConfiguration config;
	wren::VM vm(&config, nullptr);
	vm.callback_manager()
		.add_callback(true, "bitcoinity", "Bitcoinity", "ticker(_,_)", &ticker_price_bitcoinity_wren)
		.add_callback(true, "bitcoinity", "Bitcoinity", "ticker_quick()", &quick_ticker_price_bitcoinity_wren)
		.add_callback(true, "bitcoinity", "Bitcoinity", "ticker_currency(_)", &currency_ticker_price_bitcoinity_wren)
		.add_callback(true, "bitcoinity", "Bitcoinity", "ticker_exchange(_)", &exchange_ticker_price_bitcoinity_wren)
		.add_callback(false, "ticker_price", "TickerPrice", "price", wren::make_method_bindable<&TickerPrice::price>())
		.add_callback(false, "ticker_price", "TickerPrice", "currency", wren::make_method_bindable<&TickerPrice::currency>())
		.add_callback(false, "ticker_price", "TickerPrice", "is_valid", wren::make_method_bindable<&TickerPrice::is_valid>())
		.add_callback(false, "ticker_price", "TickerPrice", "exchange", wren::make_method_bindable<&TickerPrice::exchange>());

	vm.interpret("main", g_script);
	wren::call<void>(vm, {"main", "the_app"}, "start");

	return 0;
}
